#!/bin/sh
# SPDX-License-Identifier: GPL-2.0

ATOMICDIR=$(dirname $0)

. ${ATOMICDIR}/atomic-tbl.sh

#gen_cast(arg, int, atomic)
gen_cast()
{
	local arg="$1"; shift
	local int="$1"; shift
	local atomic="$1"; shift

	[ "${arg%%:*}" = "p" ] || return

	printf "($(gen_param_type "${arg}" "${int}" "${atomic}"))"
}

#gen_args_cast(int, atomic, arg...)
gen_args_cast()
{
	local int="$1"; shift
	local atomic="$1"; shift

	while [ "$#" -gt 0 ]; do
		local cast="$(gen_cast "$1" "${int}" "${atomic}")"
		local arg="$(gen_param_name "$1")"
		printf "${cast}${arg}"
		[ "$#" -gt 1 ] && printf ", "
		shift;
	done
}

#gen_proto_order_variant(meta, pfx, name, sfx, unchecked, order, arg...)
gen_proto_order_variant()
{
	local meta="$1"; shift
	local pfx="$1"; shift
	local name="$1"; shift
	local sfx="$1"; shift
	local unchecked="$1"; shift
	local order="$1"; shift

	local stemname="${pfx}${name}${sfx}"
	local atomic="atomic_long"
	local atomicname="${stemname}${unchecked}${order}"

	local ret="$(gen_ret_type "${meta}" "long")"
	local params="$(gen_params "long" "atomic_long${unchecked}" "$@")"
	local args="$(gen_args "$@")"
	local argscast_32="$(gen_args_cast "int" "atomic${unchecked}" "$@")"
	local argscast_64="$(gen_args_cast "s64" "atomic64${unchecked}" "$@")"
	local retstmt="$(gen_ret_stmt "${meta}")"

	gen_kerneldoc "raw_" "${meta}" "${pfx}" "${name}" "${sfx}" "${unchecked}" "${order}" "atomic_long" "long" "$@"

cat <<EOF
static __always_inline __turn_off_size_overflow ${ret}
raw_${atomic}_${atomicname}(${params})
{
#ifdef CONFIG_64BIT
	${retstmt}raw_atomic64_${atomicname}(${argscast_64});
#else
	${retstmt}raw_atomic_${atomicname}(${argscast_32});
#endif
}
EOF

	if [ ! -z ${unchecked} ] ; then
		printf "#define raw_${atomic}_${stemname}${order}(${args}) raw_${atomic}_op_pick(${atomic},${stemname},${order},v)(${args})\n"
	fi
	printf "\n"
}

cat << EOF
// SPDX-License-Identifier: GPL-2.0
// Copyright (C) 2009  PaX Team <pageexec@freemail.hu>

// Generated by $0
// DO NOT MODIFY THIS FILE DIRECTLY

#ifndef _LINUX_ATOMIC_LONG_H
#define _LINUX_ATOMIC_LONG_H

#include <linux/compiler.h>
#include <asm/types.h>

#if !defined(raw_atomic_long_op_pick)
#error Do not include this file directly, use <linux/atomic.h> instead!
#endif

#ifdef CONFIG_64BIT
typedef atomic64_t atomic_long_t;

#ifdef CONFIG_PAX_REFCOUNT
typedef atomic64_unchecked_t atomic_long_unchecked_t;
#else
typedef atomic64_t atomic_long_unchecked_t;
#endif

#define ATOMIC_LONG_INIT(i)		ATOMIC64_INIT(i)
#define atomic_long_cond_read_acquire	atomic64_cond_read_acquire
#define atomic_long_cond_read_relaxed	atomic64_cond_read_relaxed
#else
typedef atomic_t atomic_long_t;

#ifdef CONFIG_PAX_REFCOUNT
typedef atomic_unchecked_t atomic_long_unchecked_t;
#else
typedef atomic_t atomic_long_unchecked_t;
#endif

#define ATOMIC_LONG_INIT(i)		ATOMIC_INIT(i)
#define atomic_long_cond_read_acquire	atomic_cond_read_acquire
#define atomic_long_cond_read_relaxed	atomic_cond_read_relaxed
#endif

EOF

grep '^[au]' "$1" | while read kind name meta args; do
	gen_proto "${kind}" "${meta}" "${name}" ${args}
done

cat <<EOF
#endif /* _LINUX_ATOMIC_LONG_H */
EOF
