/* SPDX-License-Identifier: GPL-2.0 */
#ifndef GCC_COMMON_H_INCLUDED
#define GCC_COMMON_H_INCLUDED

#include "bversion.h"

#if BUILDING_GCC_VERSION >= 8000
#define IN_TARGET_CODE
#endif

#if BUILDING_GCC_VERSION >= 6000
#include <algorithm>
#include "gcc-plugin.h"
#else
#include "plugin.h"
#endif
#include "plugin-version.h"

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "line-map.h"
#include "input.h"
#include "tree.h"
#include "vec.h"

#include "tree-inline.h"
#include "version.h"
#include "rtl.h"

#if BUILDING_GCC_VERSION >= 7000
#include "memmodel.h"
#endif

#include "tm_p.h"
#include "flags.h"
//#include "insn-attr.h"
//#include "insn-codes.h"
//#include "insn-config.h"
//#include "insn-flags.h"
#include "hard-reg-set.h"
//#include "recog.h"
#include "output.h"
#include "except.h"
#include "function.h"
#include "toplev.h"

#include "expr.h"
#include "explow.h"

#include "basic-block.h"
#include "intl.h"
#include "ggc.h"

#include "graph.h"
#include "regs.h"

#include "timevar.h"
#if BUILDING_GCC_VERSION < 10000
#include "params.h"
#endif

#include "hash-map.h"

#include "emit-rtl.h"
//#include "reload.h"
//#include "ira.h"
//#include "dwarf2asm.h"
#include "debug.h"
#include "target.h"
#include "langhooks.h"
#include "cfgloop.h"

#include "langhooks-def.h"
#include "targhooks.h"
#include "tree-scalar-evolution.h"

#include "tree-ssa-loop-niter.h"
#include "gimple-builder.h"
#include "calls.h"

//#include "hosthooks.h"
#include "cgraph.h"
#include "opts.h"
//#include "coverage.h"
//#include "value-prof.h"

#include "tree-pretty-print.h"
#include "gimple-pretty-print.h"

#include "c-tree.h"
#include "c-family/c-common.h"
#include "c-family/c-pragma.h"

#include "tree-cfgcleanup.h"
#include "tree-ssa-operands.h"
#include "tree-into-ssa.h"
#include "tree-dfa.h"
// partition.h is not installed...
//#include "tree-ssa-live.h"

#include "is-a.h"

#include "diagnostic.h"
//#include "tree-diagnostic.h"
#include "tree-dump.h"
#include "tree-pass.h"

#include "pass_manager.h"
#include "df.h"

#include "predict.h"

#include "stringpool.h"
#include "attribs.h"
#include "varasm.h"
#include "stor-layout.h"
#include "internal-fn.h"
#include "gimple-expr.h"
//#include "diagnostic-color.h"
#include "context.h"
#include "tree-ssa-alias.h"
#include "tree-ssa.h"

#if BUILDING_GCC_VERSION >= 7000
#include "tree-vrp.h"
#endif

#include "print-tree.h"
#include "tree-eh.h"
#include "stmt.h"
#include "gimplify.h"

#include "gimple.h"

#include "ipa-utils.h"
#if __GNUC__ >= 12
# pragma GCC diagnostic push
# pragma GCC diagnostic ignored "-Wmissing-template-keyword"
#endif
#include "symbol-summary.h"
#if __GNUC__ >= 12
# pragma GCC diagnostic pop
#endif
#include "sreal.h"
#if BUILDING_GCC_VERSION >= 14000
#include "ipa-cp.h"
#endif
#include "ipa-prop.h"
#if BUILDING_GCC_VERSION >= 8000
#include "ipa-fnsummary.h"
#endif
#include "ipa-inline.h"

#include "tree-ssanames.h"
#include "tree-ssa-loop.h"
#include "tree-ssa-loop-manip.h"
#include "tree-phinodes.h"
#include "tree-cfg.h"
#include "gimple-iterator.h"
#include "gimple-fold.h"
#include "gimple-ssa.h"
#include "gimple-walk.h"
#include "ssa-iterators.h"
#include "gimplify-me.h"

//#include "lto/lto.h"
//#include "data-streamer.h"
//#include "streamer-hooks.h"
//#include "lto-compress.h"
//#include "lto-section-names.h"
#include "builtins.h"
#include "domwalk.h"

#if BUILDING_GCC_VERSION >= 6000
typedef gimple *gimple_stmt_ptr;
typedef const gimple *const_gimple_stmt_ptr;
#else
typedef gimple gimple_stmt_ptr;
typedef const_gimple const_gimple_stmt_ptr;
#endif

#include "tree-ssa-propagate.h"

#include "asan.h"

#if BUILDING_GCC_VERSION >= 8000
#include "profile-count.h"
#include "poly-int.h"
#include "tree-ssa-sccvn.h"
#endif

#if BUILDING_GCC_VERSION >= 11000
#include "symtab-clones.h"
#include "symtab-thunks.h"
#include "gimple-range.h"
#include "vr-values.h"
#endif

#if BUILDING_GCC_VERSION >= 12000
#include "gimple-range-path.h"
#endif

#include "dbgcnt.h"

/* missing from basic_block.h... */
void debug_dominance_info(enum cdi_direction dir);
void debug_dominance_tree(enum cdi_direction dir, basic_block root);

/* since tree-ssa-live.h is not usable due to partition.h not being installed... */
void dump_scope_blocks(FILE *file, int flags);
void debug_scope_block(tree scope, int flags);
void debug_scope_blocks(int flags);

// c-tree.h
tree build_c_cast(location_t, tree, tree);

// STATIC_ASSERT() as of 81a12b7606c5 ("Use static_assert for STATIC_ASSERT for C++11 onwards")
#ifndef STATIC_ASSERT
/* static_assert (COND, MESSAGE) is available in C++11 onwards.  */
#if __cplusplus >= 201103L
#define STATIC_ASSERT(X) \
	static_assert ((X), #X)
#else
#define STATIC_ASSERT(X) \
	typedef int assertion1[(X) ? 1 : -1] ATTRIBUTE_UNUSED
#endif
#endif

// libcpp/include/mkdeps.h isn't installed for plugins
#if BUILDING_GCC_VERSION < 10000
void deps_add_dep(struct deps *, const char *);
#else
void deps_add_dep(class mkdeps *, const char *);
#endif

#ifndef __unused
#define __unused __attribute__((unused))
#endif

#ifndef __visible
#ifdef _WIN32
#define __visible __declspec(dllexport)
#else
#define __visible __attribute__((visibility("default")))
#endif
#endif

#define __weak __attribute__((weak)) __visible

// A value high enough to not clash with any of gcc's internal builtins
#if BUILDING_GCC_VERSION < 10000
// extended to 12 bits in c1b7a563e061 ("AVX-512. 81.1/n. Extend `function_code' field `tree_var_decl' structure.")
#define BUILT_IN_PLUGINS_START	0xfc0	// 63 custom builtins max!
#else
// finally 32 bit since 55f863c4d694 ("Make function_code a 32-bit field")
#define BUILT_IN_PLUGINS_START	0xc0ffee
#endif

#define BUILT_IN_CANARY_VALUE	(BUILT_IN_PLUGINS_START + 0)
#define BUILT_IN_CANARY_COPY	(BUILT_IN_PLUGINS_START + 1)
#define BUILT_IN_MUTABLE_P	(BUILT_IN_PLUGINS_START + 2)

#define DECL_NAME_POINTER(node) IDENTIFIER_POINTER(DECL_NAME(node))
#define DECL_NAME_LENGTH(node)  IDENTIFIER_LENGTH(DECL_NAME(node))
#define TYPE_NAME_POINTER(node) IDENTIFIER_POINTER(type_name(node))
#define TYPE_NAME_LENGTH(node)  IDENTIFIER_LENGTH(type_name(node))

#define FUNC_OR_METHOD_POINTER_TYPE_P(TYPE) (POINTER_TYPE_P(TYPE) && FUNC_OR_METHOD_TYPE_P(TREE_TYPE(TYPE)))

/*
 * Helpers to process types just once.
 *
 * Meant to be used in a PLUGIN_FINISH_TYPE callback.
 * Cleanup should be done in PLUGIN_ALL_PASSES_START.
 */

#define TYPEMAP(name)	bitmap name

static inline bitmap typemap_init(void)
{
	bitmap_obstack_initialize(NULL);
	return BITMAP_ALLOC(NULL);
}

static inline void typemap_add(bitmap typemap, const_tree type)
{
	gcc_assert(typemap);
	gcc_assert(TYPE_P(type));
	bitmap_set_bit(typemap, TYPE_UID(type));
}

static inline void typemap_del(bitmap typemap, const_tree type)
{
	gcc_assert(typemap);
	gcc_assert(TYPE_P(type));
	bitmap_clear_bit(typemap, TYPE_UID(type));
}

static inline bool typemap_has(const bitmap typemap, const_tree type)
{
	gcc_assert(typemap);
	gcc_assert(TYPE_P(type));
	return bitmap_bit_p(typemap, TYPE_UID(type));
}

static inline void typemap_free(bitmap *typemap)
{
	gcc_assert(typemap);
	if (!*typemap)
		return;

	BITMAP_FREE(*typemap);
	bitmap_obstack_release(NULL);
	*typemap = NULL;
}

tree build_const_char_string(int len, const char *str);

static inline bool plugin_version_check(struct plugin_gcc_version *v1, struct plugin_gcc_version *v2)
{
#ifdef RELAXED_VERSION_CHECK
	return !strcmp(v1->basever, v2->basever);
#else
	return plugin_default_version_check(v1, v2);
#endif
}

#define plugin_default_version_check plugin_version_check

static inline void error_gcc_version(struct plugin_gcc_version *version)
{
#ifdef RELAXED_VERSION_CHECK
	if (strcmp(gcc_version.basever, version->basever))
		error(G_("incompatible gcc version: plugin needs %s but has %s"),
		      gcc_version.basever, version->basever);
#else
	if (strcmp(gcc_version.basever, version->basever) ||
	    strcmp(gcc_version.datestamp, version->datestamp) ||
	    strcmp(gcc_version.devphase, version->devphase) ||
	    strcmp(gcc_version.revision, version->revision))
		error(G_("incompatible gcc version: plugin needs %s %s %s %s but has %s %s %s %s.\n"
			 "If this was intentional or expected, consider enabling "
			 "CONFIG_GCC_RELAXED_VERSION_CHECK after reviewing its kernel config help."),
		      gcc_version.basever, gcc_version.datestamp, gcc_version.devphase, gcc_version.revision,
		      version->basever, version->datestamp, version->devphase, version->revision);
	else
		error(G_("incompatible gcc version: plugin needs a configure string of:\n%s\nbut has:\n%s\n"
			 "If this was intentional or expected, consider enabling "
			 "CONFIG_GCC_RELAXED_VERSION_CHECK after reviewing its kernel config help."),
		      gcc_version.configuration_arguments, version->configuration_arguments);
#endif
}

void add_type_attr(tree type, const char *attr, tree args);
void update_type_attr(tree type, const char *attr, tree args);

static inline bool plugin_register_event(const char *plugin_name, const char *event_name, plugin_callback_func callback, void *user_data)
{
	int event;

	event = get_named_event_id(event_name, NO_INSERT);
	if (event != -1)
		return false;

	event = get_named_event_id(event_name, INSERT);
	if (event < PLUGIN_EVENT_FIRST_DYNAMIC)
		return false;

	register_callback(plugin_name, event, callback, user_data);
	return true;
}

static inline bool plugin_send_event(const char *event_name, void *data)
{
	int event;

	event = get_named_event_id(event_name, NO_INSERT);
	if (event == -1)
		return false;

	invoke_plugin_callbacks(event, data);
	return true;
}

// TODO: use TYPE_IDENTIFIER instead?
static inline const_tree type_name(const_tree type)
{
	const_tree name;

	name = TYPE_NAME(type);
	if (!name)
		return NULL_TREE;

	switch (TREE_CODE(name)) {
	case IDENTIFIER_NODE:
		return name;

	case TYPE_DECL:
		gcc_assert(DECL_NAME(name));
		return DECL_NAME(name);

	default:
		gcc_unreachable();
	}
}

basic_block plugin_split_edge(edge e);

#define PASS_INFO(NAME, REF, ID, POS)		\
struct register_pass_info NAME##_pass_info = {	\
	.pass = make_pass_##NAME(g),		\
	.reference_pass_name = REF,		\
	.ref_pass_instance_number = ID,		\
	.pos_op = POS,				\
}

#if BUILDING_GCC_VERSION < 8000
#define ATTRIBUTE_SPEC_INIT(NAME, MIN, MAX, DECL, TYPE, FUNC, HANDLER, IDENTITY, EXCLUDE) \
{ \
	.name				= NAME,		\
	.min_length			= MIN,		\
	.max_length			= MAX,		\
	.decl_required			= DECL,		\
	.type_required			= TYPE,		\
	.function_type_required		= FUNC,		\
	.handler			= HANDLER,	\
	.affects_type_identity		= IDENTITY,	\
}
#else
#define ATTRIBUTE_SPEC_INIT(NAME, MIN, MAX, DECL, TYPE, FUNC, HANDLER, IDENTITY, EXCLUDE) \
{ \
	.name				= NAME,		\
	.min_length			= MIN,		\
	.max_length			= MAX,		\
	.decl_required			= DECL,		\
	.type_required			= TYPE,		\
	.function_type_required		= FUNC,		\
	.affects_type_identity		= IDENTITY,	\
	.handler			= HANDLER,	\
	.exclude			= EXCLUDE,	\
}
#endif

#define INIT_GGC_ROOT_TAB_TREE(NAME)	\
{ \
	.base = &NAME,			\
	.nelt = 1,			\
	.stride = sizeof(NAME),		\
	.cb = &gt_ggc_mx_tree_node,	\
	.pchw = &gt_pch_nx_tree_node	\
}

#define INIT_GGC_ROOT_TAB_TREES(NAME, FIELD) \
{ \
	.base = &NAME[0].FIELD,		\
	.nelt = ARRAY_SIZE(NAME),	\
	.stride = sizeof(NAME[0]),	\
	.cb = &gt_ggc_mx_tree_node,	\
	.pchw = &gt_pch_nx_tree_node	\
}

#define INIT_GGC_ROOT_TAB_RTX(NAME)	\
{ \
	.base = &NAME,			\
	.nelt = 1,			\
	.stride = sizeof(NAME),		\
	.cb = &gt_ggc_mx_rtx_def,	\
	.pchw = &gt_pch_nx_rtx_def	\
}


#define VEC_index(T, V, I) ((*(V))[(I)])
#define VEC_replace(T, V, I, O) (*(V))[(I)] = (O)
#define VEC_safe_push(T, A, V, O) vec_safe_push((V), (O))

#define SET_INSN_LOCATION(INSN, LOC) INSN_LOCATION(INSN) = (LOC)

#define EXECUTE_IF_SET_IN_SBITMAP(BITMAP, MIN, BITNUM, ITER)		\
	for (bmp_iter_set_init(&(ITER), (BITMAP), (MIN), &(BITNUM));	\
		bmp_iter_set (&(ITER), &(BITNUM));			\
		bmp_iter_next (&(ITER), &(BITNUM)))

static inline SBITMAP_ELT_TYPE sbitmap_bit_p(const_sbitmap map, unsigned int bitno)
{
	return bitmap_bit_p(map, bitno);
}

static inline void sbitmap_set_bit(sbitmap map, unsigned int bitno)
{
	bitmap_set_bit(map, bitno);
}

static inline void sbitmap_clear_bit(sbitmap map, unsigned int bitno)
{
	bitmap_clear_bit(map, bitno);
}

static inline void sbitmap_clear(sbitmap map)
{
	bitmap_clear(map);
}

static inline void sbitmap_ones(sbitmap map)
{
	bitmap_ones(map);
}

static inline int sbitmap_first_set_bit(const_sbitmap map)
{
	return bitmap_first_set_bit(map);
}

static inline void debug_sbitmap(const_sbitmap map)
{
	debug_bitmap(map);
}

static inline bool sbitmap_empty_p(const_sbitmap map)
{
	return bitmap_empty_p(map);
}

static inline unsigned int sbitmap_count_bits(const_sbitmap map)
{
	unsigned int count = 0;

	for (unsigned int i = 0; i < map->size; i++)
		if (map->elms[i]) {
#if HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONG
		count += __builtin_popcountl(map->elms[i]);
#elif HOST_BITS_PER_WIDEST_FAST_INT == HOST_BITS_PER_LONGLONG
		count += __builtin_popcountll(map->elms[i]);
#else
		count += __builtin_popcount(map->elms[i]);
#endif
		}

	return count;
}

static inline void sbitmap_copy(sbitmap dst, const_sbitmap src)
{
	memcpy(dst->elms, src->elms, sizeof(SBITMAP_ELT_TYPE) * dst->size);
}

static inline int sbitmap_first_set_bit(const sbitmap map)
{
	return bitmap_first_set_bit(map);
}

static inline int sbitmap_last_set_bit(const sbitmap map)
{
	return bitmap_last_set_bit(map);
}

static inline bool sbitmap_and(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return bitmap_and(dst, a, b);
}

static inline bool sbitmap_ior(sbitmap dst, const_sbitmap a, const_sbitmap b)
{
	return bitmap_ior(dst, a, b);
}

basic_block nearest_common_dominator_for_sbitmap(enum cdi_direction dir, const_sbitmap blocks);

#define NODE_DECL(node) (node)->decl
#define cgraph_node_name(node) (node)->name()
#define NODE_IMPLICIT_ALIAS(node) (node)->cpp_implicit_alias

static inline opt_pass *get_pass_for_id(int id)
{
	return g->get_passes()->get_pass_for_id(id);
}

//#define TREE_INT_CST_HIGH(NODE) ({ TREE_INT_CST_EXT_NUNITS(NODE) > 1 ? (unsigned HOST_WIDE_INT)TREE_INT_CST_ELT(NODE, 1) : 0; })

#define INSN_DELETED_P(insn) (insn)->deleted()

static inline gphi *gsi_gphi(gphi_iterator gsi)
{
	return gsi.phi();
}

static inline const char *get_decl_section_name(const_tree decl)
{
	return DECL_SECTION_NAME(decl);
}

#define set_decl_section_name(_node, _value)						\
do {											\
	tree node = (_node);								\
	const char *value = (_value);							\
											\
	DECL_ATTRIBUTES(node) = remove_attribute("section", DECL_ATTRIBUTES(node));	\
	set_decl_section_name(node, value);						\
	if (value) {									\
		tree section = build_const_char_string(strlen(value) + 1, value);	\
											\
		section = tree_cons(NULL_TREE, section, NULL_TREE);			\
		DECL_ATTRIBUTES(node) = tree_cons(get_identifier("section"), section, DECL_ATTRIBUTES(node)); \
	}										\
} while (0)

/* symtab/cgraph related */
#define debug_cgraph_node(node) (node)->debug()
#define cgraph_get_node(decl) cgraph_node::get(decl)
#define cgraph_get_create_node(decl) cgraph_node::get_create(decl)
#define cgraph_create_node(decl) cgraph_node::create(decl)
#define cgraph_record_stmt_references(node, stmt) (node)->record_stmt_references(stmt)
#define cgraph_node_for_asm(asmname) cgraph_node::get_for_asmname(asmname)
#define cgraph_n_nodes symtab->cgraph_count
#define cgraph_max_uid symtab->cgraph_max_uid
#define cgraph_edge_recursive_p(edge) edge->recursive_p()
#define varpool_get_node(decl) varpool_node::get(decl)
#define varpool_get_for_asmname(asmname) varpool_node::get_for_asmname(asmname)
#define varpool_assemble_decl(node) (node)->assemble_decl()
#define dump_varpool_node(file, node) (node)->dump(file)

typedef struct cgraph_node *cgraph_node_ptr;
typedef struct cgraph_edge *cgraph_edge_p;
typedef struct varpool_node *varpool_node_ptr;

void clone_late_inlined_nodes(cgraph_edge_p edge);

static inline void change_decl_assembler_name(tree decl, tree name)
{
	symtab->change_decl_assembler_name(decl, name);
}

static inline void varpool_finalize_decl(tree decl)
{
	varpool_node::finalize_decl(decl);
}

static inline void varpool_add_new_variable(tree decl)
{
	varpool_node::add(decl);
}

static inline unsigned int rebuild_cgraph_edges(void)
{
	return cgraph_edge::rebuild_edges();
}

static inline cgraph_node_ptr cgraph_function_node(cgraph_node_ptr node, enum availability *availability)
{
	return node->function_symbol(availability);
}

static inline cgraph_node_ptr cgraph_function_or_thunk_node(cgraph_node_ptr node, enum availability *availability = NULL)
{
	return node->ultimate_alias_target(availability);
}

static inline bool cgraph_only_called_directly_p(cgraph_node_ptr node)
{
	return node->only_called_directly_p();
}

static inline enum availability cgraph_function_body_availability(cgraph_node_ptr node)
{
	return node->get_availability();
}

static inline cgraph_node_ptr cgraph_alias_target(cgraph_node_ptr node)
{
	return node->get_alias_target();
}

static inline bool cgraph_for_node_and_aliases(cgraph_node_ptr node, bool (*callback)(cgraph_node_ptr, void *), void *data, bool include_overwritable)
{
	return node->call_for_symbol_and_aliases(callback, data, include_overwritable);
}

static inline bool cgraph_for_node_thunks_and_aliases(cgraph_node_ptr node, bool (*callback)(cgraph_node_ptr, void *), void *data, bool include_overwritable)
{
	return node->call_for_symbol_thunks_and_aliases(callback, data, include_overwritable);
}

static inline struct cgraph_node_hook_list *cgraph_add_function_insertion_hook(cgraph_node_hook hook, void *data)
{
	return symtab->add_cgraph_insertion_hook(hook, data);
}

static inline void cgraph_remove_function_insertion_hook(struct cgraph_node_hook_list *entry)
{
	symtab->remove_cgraph_insertion_hook(entry);
}

static inline struct cgraph_node_hook_list *cgraph_add_node_removal_hook(cgraph_node_hook hook, void *data)
{
	return symtab->add_cgraph_removal_hook(hook, data);
}

static inline void cgraph_remove_node_removal_hook(struct cgraph_node_hook_list *entry)
{
	symtab->remove_cgraph_removal_hook(entry);
}

static inline struct cgraph_2node_hook_list *cgraph_add_node_duplication_hook(cgraph_2node_hook hook, void *data)
{
	return symtab->add_cgraph_duplication_hook(hook, data);
}

static inline void cgraph_remove_node_duplication_hook(struct cgraph_2node_hook_list *entry)
{
	symtab->remove_cgraph_duplication_hook(entry);
}

static inline void cgraph_call_node_duplication_hooks(cgraph_node_ptr node, cgraph_node_ptr node2)
{
	symtab->call_cgraph_duplication_hooks(node, node2);
}

static inline void cgraph_call_edge_duplication_hooks(cgraph_edge_p cs1, cgraph_edge_p cs2)
{
	symtab->call_edge_duplication_hooks(cs1, cs2);
}

/* gimple related */
static inline gimple_stmt_ptr gimple_build_assign_with_ops(enum tree_code subcode, tree lhs, tree op1, tree op2 MEM_STAT_DECL)
{
	return gimple_build_assign(lhs, subcode, op1, op2 PASS_MEM_STAT);
}

template <>
template <>
inline bool is_a_helper<rtx_insn *>::test(rtx_insn *insn __unused)
{
	return true;
}

static inline gasm *as_a_gasm(gimple_stmt_ptr stmt)
{
	return as_a<gasm *>(stmt);
}

static inline const gasm *as_a_const_gasm(const_gimple_stmt_ptr stmt)
{
	return as_a<const gasm *>(stmt);
}

static inline gassign *as_a_gassign(gimple_stmt_ptr stmt)
{
	return as_a<gassign *>(stmt);
}

static inline const gassign *as_a_const_gassign(const_gimple_stmt_ptr stmt)
{
	return as_a<const gassign *>(stmt);
}

static inline gcall *as_a_gcall(gimple_stmt_ptr stmt)
{
	return as_a<gcall *>(stmt);
}

static inline const gcall *as_a_const_gcall(const_gimple_stmt_ptr stmt)
{
	return as_a<const gcall *>(stmt);
}

static inline gcond *as_a_gcond(gimple_stmt_ptr stmt)
{
	return as_a<gcond *>(stmt);
}

static inline const gcond *as_a_const_gcond(const_gimple_stmt_ptr stmt)
{
	return as_a<const gcond *>(stmt);
}

static inline ggoto *as_a_ggoto(gimple_stmt_ptr stmt)
{
	return as_a<ggoto *>(stmt);
}

static inline const ggoto *as_a_const_ggoto(const_gimple_stmt_ptr stmt)
{
	return as_a<const ggoto *>(stmt);
}

static inline gphi *as_a_gphi(gimple_stmt_ptr stmt)
{
	return as_a<gphi *>(stmt);
}

static inline const gphi *as_a_const_gphi(const_gimple_stmt_ptr stmt)
{
	return as_a<const gphi *>(stmt);
}

static inline greturn *as_a_greturn(gimple_stmt_ptr stmt)
{
	return as_a<greturn *>(stmt);
}

static inline const greturn *as_a_const_greturn(const_gimple_stmt_ptr stmt)
{
	return as_a<const greturn *>(stmt);
}

static inline gswitch *as_a_gswitch(gimple_stmt_ptr stmt)
{
	return as_a<gswitch *>(stmt);
}

static inline const glabel *as_a_const_glabel(const_gimple_stmt_ptr stmt)
{
	return as_a<const glabel *>(stmt);
}

static inline const gswitch *as_a_const_gswitch(const_gimple_stmt_ptr stmt)
{
	return as_a<const gswitch *>(stmt);
}

/* pointer/hash set related */
#define HASH_SET(hash, type)						\
typedef struct hash_set<type> hash##_set;				\
									\
static inline hash##_set* hash##_pointer_set_create(void)		\
{									\
	return new hash##_set;						\
}									\
									\
static inline hash##_set* hash##_pointer_set_create_ggc(size_t n = 13)	\
{									\
	return hash##_set::create_ggc(n);				\
}									\
									\
static inline void pointer_set_destroy(hash##_set *set)			\
{									\
	delete set;							\
}									\
									\
static inline bool pointer_set_insert(hash##_set *set, type ptr)	\
{									\
	return set->add(ptr);						\
}									\
									\
static inline bool pointer_set_contains(hash##_set *set, type ptr)	\
{									\
	return set->contains(ptr);					\
}

HASH_SET(gimple, gimple_stmt_ptr)
HASH_SET(const_gimple, const_gimple_stmt_ptr)
HASH_SET(tree, tree)
HASH_SET(const_tree, const_tree)
HASH_SET(cgraph, cgraph_node_ptr)

#define pointer_set_traverse(set, fn, data) set->traverse<void *, fn>(data)
#define pointer_set_elements(set) set->elements()

/* pointer/hash map related */
#define HASH_MAP(hash, key, value)					\
typedef struct hash_map<key, value> hash##_map;				\
									\
static inline hash##_map *hash##_map_create(void)			\
{									\
	return new hash##_map;						\
}									\
									\
static inline hash##_map *hash##_map_create_ggc(size_t n = 13)		\
{									\
	return hash##_map::create_ggc(n);				\
}									\
									\
static inline void hash##_map_destroy(hash##_map *map)			\
{									\
	delete map;							\
}									\
									\
static inline bool hash##_map_insert(hash##_map *map, key ptr1, value ptr2)\
{									\
	return map->put(ptr1, ptr2);					\
}									\
									\
static inline bool hash##_map_contains(hash##_map *map, key ptr)	\
{									\
	return !!map->get(ptr);						\
}									\
									\
static inline value *hash##_map_get(hash##_map *map, key ptr)		\
{									\
	return map->get(ptr);						\
}

HASH_MAP(tree_tree, tree, tree)

#define tree_tree_map_traverse(map, fn, data) map->traverse<void *, fn>(data)
#define tree_tree_map_elements(map) map->elements()

/* IPA/LTO related */
#define ipa_ref_list_referring_iterate(L, I, P)	\
	(L)->referring.iterate((I), &(P))
#define ipa_ref_list_reference_iterate(L, I, P)	\
	(L)->reference.iterate((I), &(P))

static inline cgraph_node_ptr ipa_ref_referring_cgraph_node(struct ipa_ref *ref)
{
	return dyn_cast<cgraph_node_ptr>(ref->referring);
}

static inline varpool_node_ptr ipa_ref_referring_varpool_node(struct ipa_ref *ref)
{
	return dyn_cast<varpool_node_ptr>(ref->referring);
}

static inline void ipa_remove_stmt_references(symtab_node *referring_node, gimple_stmt_ptr stmt)
{
	referring_node->remove_stmt_references(stmt);
}

#if BUILDING_GCC_VERSION < 6000
/* gimple related */
template <>
template <>
inline bool is_a_helper<const gassign *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_ASSIGN;
}

template <>
template <>
inline bool is_a_helper<const gcond *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_COND;
}

#define CALL_USED_REGISTERS_MASK(IS_64BIT_MS_ABI) ((IS_64BIT_MS_ABI) ? (1 << 3) : TARGET_64BIT ? (1 << 2) : (1 << 1))

#define get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep, keep_aligning)	\
	({ (void)(preversep); get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, pvolatilep, keep_aligning); })
#define gen_rtx_set(ARG0, ARG1) gen_rtx_SET(VOIDmode, (ARG0), (ARG1))

#ifndef LEGACY_INT_REG_P
#define LEGACY_INT_REG_P(X) (REG_P(X) && LEGACY_INT_REGNO_P(REGNO(X)))
#endif
#ifndef LEGACY_INT_REGNO_P
#define LEGACY_INT_REGNO_P(N) (IN_RANGE((N), AX_REG, SP_REG))
#endif

#ifndef FIRST_INT_REG
#define FIRST_INT_REG		AX_REG
#endif
#ifndef LAST_INT_REG
#define LAST_INT_REG		SP_REG
#endif

#define FUNC_OR_METHOD_TYPE_P(NODE) (TREE_CODE(NODE) == FUNCTION_TYPE || TREE_CODE(NODE) == METHOD_TYPE)
#endif

#if BUILDING_GCC_VERSION >= 6000
#define gen_rtx_set(ARG0, ARG1) gen_rtx_SET((ARG0), (ARG1))
#endif

#if BUILDING_GCC_VERSION < 7000
static inline source_location expansion_point_location(source_location location)
{
	return linemap_resolve_location(line_table, location, LRK_MACRO_EXPANSION_POINT, NULL);
}

#define SET_DECL_ALIGN(n, a)	({ DECL_ALIGN(n) = (a); })
#define SET_DECL_MODE(n, m)	({ DECL_MODE(n) = (m); })
#define FOR_EACH_SSA_NAME(I, VAR, FN)	for ((I) = 1; (I) < num_ssa_names; (I)++) if (((VAR) = ssa_name(I)) != NULL_TREE)

static inline unsigned HOST_WIDE_INT least_bit_hwi(unsigned HOST_WIDE_INT x)
{
	return x & -x;
}

static inline bool pow2_or_zerop(unsigned HOST_WIDE_INT x)
{
	return least_bit_hwi(x) == x;
}

static inline bool pow2p_hwi(unsigned HOST_WIDE_INT x)
{
	return x && pow2_or_zerop(x);
}
#endif

#if BUILDING_GCC_VERSION >= 7000
#define get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep, keep_aligning)	\
	get_inner_reference(exp, pbitsize, pbitpos, poffset, pmode, punsignedp, preversep, pvolatilep)
#endif

#if BUILDING_GCC_VERSION < 8000
#define ALLOCA_FUNCTION_CODE_P(FCODE)	\
	((FCODE) == BUILT_IN_ALLOCA || (FCODE) == BUILT_IN_ALLOCA_WITH_ALIGN)

static inline void remove_call_arg_locations(rtx_insn *insn)
{
	rtx_insn *next;

	for (next = NEXT_INSN(insn); next; next = NEXT_INSN(next)) {
		rtx_insn *tmp;

		if (GET_CODE(next) == BARRIER)
			continue;
		if (GET_CODE(next) != NOTE)
			break;
		if (NOTE_KIND(next) != NOTE_INSN_CALL_ARG_LOCATION)
			continue;

		tmp = PREV_INSN(next);
		delete_insn(next);
		next = tmp;
	}
}

#define REG_BR_PROB_NEVER		0
#define REG_BR_PROB_VERY_UNLIKELY	PROB_VERY_UNLIKELY
#define REG_BR_PROB_UNLIKELY		PROB_UNLIKELY
#define REG_BR_PROB_EVEN		PROB_EVEN
#define REG_BR_PROB_LIKELY		PROB_LIKELY
#define REG_BR_PROB_VERY_LIKELY		PROB_VERY_LIKELY
#define REG_BR_PROB_ALWAYS		PROB_ALWAYS

static inline cgraph_edge_p cgraph_create_edge(cgraph_node_ptr caller, cgraph_node_ptr callee, gcall *call_stmt, gcov_type count, int freq)
{
	cgraph_edge_p edge = caller->create_edge(callee, call_stmt, count, freq);
	cgraph_record_stmt_references(caller, call_stmt);
	return edge;
}

#define clone_inlined_nodes(edge, duplicate, update_original, overall_size) \
	clone_inlined_nodes((edge), (duplicate), (update_original), (overall_size), (edge)->frequency)

typedef unsigned short poly_uint16_pod;
typedef HOST_WIDE_INT poly_int64_pod;
typedef unsigned HOST_WIDE_INT poly_uint64_pod;
typedef offset_int poly_offset_int_pod;
typedef wide_int poly_wide_int_pod;
typedef widest_int poly_widest_int_pod;

typedef unsigned short poly_uint16;
typedef HOST_WIDE_INT poly_int64;
typedef unsigned HOST_WIDE_INT poly_uint64;
typedef offset_int poly_offset_int;
typedef wide_int poly_wide_int;
typedef wide_int_ref poly_wide_int_ref;
typedef widest_int poly_widest_int;

#define RETURN_FROM_IMM_USE_STMT(ITER, VAL) {	\
	end_imm_use_stmt_traverse(&(ITER));	\
	return (VAL);				\
}

static inline bool multiple_p(poly_int64_pod a, HOST_WIDE_INT b)
{
	return a % b == 0;
}

static inline poly_int64_pod exact_div(poly_int64_pod a, HOST_WIDE_INT b)
{
	return a / b;
}

static inline void print_dec(poly_int64_pod a, FILE *f)
{
	fprintf(f, "%ld", a);
}

static inline void print_gimple_stmt(FILE *file, const_gimple_stmt_ptr g, int spc, int flags)
{
	print_gimple_stmt(file, CONST_CAST_GIMPLE(g), spc, flags);
}

static inline void print_gimple_expr(FILE *file, const_gimple_stmt_ptr g, int spc, int flags)
{
	print_gimple_expr(file, CONST_CAST_GIMPLE(g), spc, flags);
}

#define edge_count(e) (e)->count

#define E_BLKmode BLKmode
#define E_BImode BImode
#define E_QImode QImode
#define E_HImode HImode
#define E_SImode SImode
#define E_DImode DImode
#define E_TImode TImode
#define E_OImode OImode
#define E_XImode XImode
#define E_BND32mode BND32mode
#define E_BND64mode BND64mode
#define E_QQmode QQmode
#define E_HQmode HQmode
#define E_SQmode SQmode
#define E_DQmode DQmode
#define E_TQmode TQmode
#define E_UQQmode UQQmode
#define E_UHQmode UHQmode
#define E_USQmode USQmode
#define E_UDQmode UDQmode
#define E_UTQmode UTQmode
#define E_HAmode HAmode
#define E_SAmode SAmode
#define E_DAmode DAmode
#define E_TAmode TAmode
#define E_UHAmode UHAmode
#define E_USAmode USAmode
#define E_UDAmode UDAmode
#define E_UTAmode UTAmode
#define E_SFmode SFmode
#define E_DFmode DFmode
#define E_XFmode XFmode
#define E_TFmode TFmode
#define E_SDmode SDmode
#define E_DDmode DDmode
#define E_TDmode TDmode

static inline void replace_call_with_value(gimple_stmt_iterator *gsi, tree val)
{
	gimple_stmt_ptr repl, stmt = gsi_stmt(*gsi);
	tree vdef, lhs = gimple_call_lhs(stmt);

	if (lhs) {
		if (!useless_type_conversion_p(TREE_TYPE(lhs), TREE_TYPE(val)))
			val = fold_convert(TREE_TYPE(lhs), val);
		repl = gimple_build_assign(lhs, val);
	} else
		repl = gimple_build_nop();

	vdef = gimple_vdef(stmt);
	if (vdef && TREE_CODE(vdef) == SSA_NAME) {
		unlink_stmt_vdef(stmt);
		release_ssa_name(vdef);
	}

	gsi_replace(gsi, repl, false);
}

static inline bool id_equal(const_tree id, const char *str)
{
	return !strcmp(IDENTIFIER_POINTER(id), str);
}

#define gen_known_cond_type(cond, op, type)				\
static inline bool known_##cond##_##type(const type& a, const type& b)	\
{									\
	return a op b;							\
}
#else
#define gen_known_cond_type(cond, op, type)				\
static inline bool known_##cond##_##type(const type& a, const type& b)	\
{									\
	return known_##cond(a, b);					\
}
#endif

#define gen_known_type(type)			\
	gen_known_cond_type(eq, ==, type)	\
	gen_known_cond_type(ne, !=, type)	\
	gen_known_cond_type(le, <=, type)	\
	gen_known_cond_type(lt, <,  type)	\
	gen_known_cond_type(ge, >=, type)	\
	gen_known_cond_type(gt, >,  type)

gen_known_type(poly_uint16)
gen_known_type(poly_int64)

#undef gen_known_type_cond
#undef gen_known_type

#if BUILDING_GCC_VERSION >= 8000
#define empty_string ""
#define REG_BR_PROB_NEVER		profile_probability::never()
#define REG_BR_PROB_VERY_UNLIKELY	profile_probability::very_unlikely()
#define REG_BR_PROB_UNLIKELY		profile_probability::unlikely()
#define REG_BR_PROB_EVEN		profile_probability::even()
#define REG_BR_PROB_LIKELY		profile_probability::likely()
#define REG_BR_PROB_VERY_LIKELY		profile_probability::very_likely()
#define REG_BR_PROB_ALWAYS		profile_probability::always()
#define edge_count(e) (e)->count()

static inline cgraph_edge_p cgraph_create_edge(cgraph_node_ptr caller, cgraph_node_ptr callee, gcall *call_stmt, profile_count count, int freq)
{
	cgraph_edge_p edge = caller->create_edge(callee, call_stmt, count);
	cgraph_record_stmt_references(caller, call_stmt);
	return edge;
}

#if 0
namespace wi {
	template <>
	struct int_traits <poly_uint16> {
//		static const enum precision_type precision_type = FLEXIBLE_PRECISION;
		static const enum precision_type precision_type = VAR_PRECISION;
		static const bool host_dependent_precision = true;
		static const bool is_sign_extended = true;
		static unsigned int get_precision(const poly_uint16& x) { return 16; }
		static wi::storage_ref decompose(HOST_WIDE_INT *scratch, unsigned int precision, const poly_uint16 &x);
	};
}
#endif

static inline void remove_call_arg_locations(rtx_insn *insn) {}

static inline void print_gimple_stmt(FILE *file, const_gimple_stmt_ptr g, int spc, dump_flags_t flags = TDF_NONE)
{
	print_gimple_stmt(file, CONST_CAST_GIMPLE(g), spc, flags);
}

static inline void print_gimple_expr(FILE *file, const_gimple_stmt_ptr g, int spc, dump_flags_t flags = TDF_NONE)
{
	print_gimple_expr(file, CONST_CAST_GIMPLE(g), spc, flags);
}

// hack to work around a mismatched prototype in gcc headers...
void debug_function(tree fn, dump_flags_t flags);
extern inline void debug_function(tree fn, int flags)
{
	debug_function(fn, (dump_flags_t)flags);
}
#endif

#if BUILDING_GCC_VERSION < 9000
#define NODE_UID(node) (node)->uid
#define fndecl_built_in_p(fndecl) DECL_BUILT_IN(fndecl)
#define fndecl_built_in_class_p(fndecl, klass) (DECL_BUILT_IN(fndecl) && DECL_BUILT_IN_CLASS(fndecl) == (klass))
#define stmt_can_throw_internal(fun, stmt) stmt_can_throw_internal(stmt)
#endif

#if BUILDING_GCC_VERSION >= 9000
#define fndecl_built_in_class_p(fndecl, klass) fndecl_built_in_p((fndecl), (klass))
#define align_functions_log align_functions.levels[0].log
#define NODE_UID(node) (node)->get_uid()

static inline basic_block label_to_block_fn(struct function *ifun, tree dest)
{
	return label_to_block(ifun, dest);
}
#endif

#if BUILDING_GCC_VERSION >= 7000 && BUILDING_GCC_VERSION < 10000
static inline bool ipcp_get_parm_bits(tree parm, tree *value, widest_int *mask)
{
	cgraph_node_ptr cnode;
#if BUILDING_GCC_VERSION >= 9000
	ipcp_transformation *ts;
#else
	ipcp_transformation_summary *ts;
#endif
	tree p;
	int i;
	ipa_bits *bits;

	cnode = cgraph_node::get(current_function_decl);
	ts = ipcp_get_transformation_summary(cnode);
	if (!ts || vec_safe_length(ts->bits) == 0)
		return false;

	p = DECL_ARGUMENTS(current_function_decl);
	i = 0;
	while (p && p != parm) {
		i++;
		p = DECL_CHAIN(p);
	}
	gcc_assert(p);

#if BUILDING_GCC_VERSION >= 10000
	if (NODE_CLONE_INFO(cnode) && NODE_CLONE_INFO(cnode)->param_adjustments) {
		i = NODE_CLONE_INFO(cnode)->param_adjustments->get_original_index(i);
		if (i < 0)
			return false;
	}
#endif

	bits = (*ts->bits)[i];
	if (!bits)
		return false;

	*mask = bits->mask;
	*value = wide_int_to_tree(TREE_TYPE(parm), bits->value);
	return true;
}
#endif

#if BUILDING_GCC_VERSION < 10000
template <>
template <>
inline bool is_a_helper<const ggoto *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_GOTO;
}

template <>
template <>
inline bool is_a_helper<const glabel *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_LABEL;
}

template <>
template <>
inline bool is_a_helper<const greturn *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_RETURN;
}

template <>
template <>
inline bool is_a_helper<const gswitch *>::test(const_gimple_stmt_ptr gs)
{
	return gs->code == GIMPLE_SWITCH;
}

#define NODE_LOCAL(node) (&(node)->local)
#define NODE_GLOBAL(node) (&(node)->global)
#endif

#if BUILDING_GCC_VERSION >= 10000
#define NODE_LOCAL(node) (node)
#define NODE_GLOBAL(node) (node)

rtx maybe_gen_leave(machine_mode);
#endif

#if BUILDING_GCC_VERSION >= 10000 && BUILDING_GCC_VERSION < 14000
static inline rtx gen_leave(machine_mode arg0)
{
	rtx res = maybe_gen_leave(arg0);
	gcc_assert(res);
	return res;
}
#endif

#if BUILDING_GCC_VERSION < 11000
#define VEC(T, A) vec<T, va_##A>
#define NODE_CLONE_INFO(node) (&(node)->clone)
#define NODE_THUNK(node) ((node)->thunk.thunk_p)
#define DECL_IS_UNDECLARED_BUILTIN(decl) DECL_IS_BUILTIN(decl)
#define get_range_info(stmt, expr, min, max) get_range_info((expr), (min), (max))
#endif

#if BUILDING_GCC_VERSION >= 11000
#define VEC(T, A) auto_vec<T>
#define NODE_CLONE_INFO(node) (clone_info::get(node))
#define NODE_THUNK(node) ((node)->thunk)
#define BREAK_FROM_IMM_USE_STMT(iter) break
#define RETURN_FROM_IMM_USE_STMT(iter, val) return (val)

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max);

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree name, irange &vr)
{
	wide_int wmin, wmax;
	enum value_range_kind kind = get_range_info(stmt, name, &wmin, &wmax);

	if (kind == VR_VARYING)
		vr.set_varying(TREE_TYPE(name));
	else if (kind == VR_UNDEFINED)
		vr.set_undefined();
	else {
#if BUILDING_GCC_VERSION < 13000
		tree min, max;

		min = wide_int_to_tree(TREE_TYPE(name), wmin);
		max = wide_int_to_tree(TREE_TYPE(name), wmax);
		vr.set(min, max, kind);
#else
		vr.set(TREE_TYPE(name), wmin, wmax, kind);
#endif
	}
	return kind;
}
#endif

#if BUILDING_GCC_VERSION < 12000
static inline bool gimple_deferred_init_p(gimple_stmt_ptr stmt)
{
	return false;
}
#endif

#if BUILDING_GCC_VERSION >= 11000 && BUILDING_GCC_VERSION < 12000
static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max)
{
	gimple_ranger rq;
	value_range vr;

	gcc_assert(!POINTER_TYPE_P(TREE_TYPE(expr)));
	gcc_assert(min && max);

	if (TREE_CODE(expr) == INTEGER_CST) {
		*min = wi::to_wide(expr);
		*max = *min;
		return VR_RANGE;
	}
	if (TREE_CODE(expr) != SSA_NAME)
		return VR_VARYING;

	if (!rq.range_of_expr(vr, expr, stmt))
		return VR_VARYING;
	if (vr.undefined_p() || vr.varying_p())
		return VR_VARYING;
	gcc_assert(!vr.symbolic_p());

	*min = wi::to_wide(vr.min());
	*max = wi::to_wide(vr.max());
	return vr.kind();
}
#endif

#if BUILDING_GCC_VERSION >= 12000
#ifdef TARGET_CPU_P
#define TARGET_386
#endif

#define FOR_EACH_LOOP(loop, flags) for (auto loop : loops_list(cfun, (flags)))
#define FOR_EACH_LOOP_FN(fn, loop, flags) for (auto loop : loops_list((fn), (flags)))

static inline bool gimple_deferred_init_p(gimple_stmt_ptr stmt)
{
	if (gimple_assign_single_p(stmt)) {
		tree rhs;

		rhs = gimple_assign_rhs1(stmt);
		if (TREE_CODE(rhs) != SSA_NAME)
			return false;
		stmt = SSA_NAME_DEF_STMT(rhs);
	}

	return gimple_call_internal_p(stmt, IFN_DEFERRED_INIT);
}

static inline enum value_range_kind get_range_info(gimple_stmt_ptr stmt, tree expr, wide_int *min, wide_int *max)
{
	range_query *rq;
	value_range vr;

	gcc_assert(!POINTER_TYPE_P(TREE_TYPE(expr)));
	gcc_assert(min && max);

	if (TREE_CODE(expr) == INTEGER_CST) {
		*min = wi::to_wide(expr);
		*max = *min;
		return VR_RANGE;
	}
	if (TREE_CODE(expr) != SSA_NAME)
		return VR_VARYING;

	rq = get_range_query(cfun);
	if (!rq->range_of_expr(vr, expr, stmt))
		return VR_VARYING;
	if (vr.undefined_p() || vr.varying_p())
		return VR_VARYING;
#if BUILDING_GCC_VERSION < 14000
	gcc_assert(!vr.symbolic_p());

	*min = wi::to_wide(vr.min());
	*max = wi::to_wide(vr.max());
	return vr.kind();
#else
	tree _min, _max;
	enum value_range_kind kind = get_legacy_range(vr, _min, _max);
	*min = wi::to_wide(_min);
	*max = wi::to_wide(_max);
	return kind;
#endif
}
#endif

#if BUILDING_GCC_VERSION < 13000
static inline gcall *gimple_build_builtin_unreachable(location_t loc)
{
	gcall *g = as_a_gcall(gimple_build_call(builtin_decl_implicit(BUILT_IN_UNREACHABLE), 0));
	gimple_call_set_ctrl_altering(g, true);
	gimple_set_location(g, loc);
	return g;
}
#endif

#if BUILDING_GCC_VERSION >= 13000
#ifdef TARGET_386
extern tree ix86_bf16_type_node;
#else
#define ix86_bf16_type_node NULL
#endif
#endif

#if BUILDING_GCC_VERSION < 14000
#define last_nondebug_stmt(bb) last_stmt(bb)
#endif

#if BUILDING_GCC_VERSION >= 14000
#define poly_int_pod poly_int
#define poly_int64_pod poly_int64
#endif

static inline void debug_tree(const_tree t)
{
	debug_tree(CONST_CAST_TREE(t));
}

static inline void debug_gimple_stmt(const_gimple_stmt_ptr s)
{
	debug_gimple_stmt(CONST_CAST_GIMPLE(s));
}

static inline void debug_bb(const_basic_block bb)
{
	debug_bb(CONST_CAST_BB(bb));
}

static inline location_t gimple_expansion_point_location(const_gimple_stmt_ptr g)
{
	return expansion_point_location(gimple_location(g));
}

location_t gimple_inline_point_location(const_gimple_stmt_ptr g);

static inline bool bb_any_loop(basic_block bb)
{
	return bb_loop_depth(bb) || (bb->flags & BB_IRREDUCIBLE_LOOP);
}

static inline tree get_inner_ref(tree arg)
{
	bool done = false;

	while (!done) {
		switch (TREE_CODE(arg)) {
		default:
			done = true;
			break;

		case MEM_REF:
			done = true;
			break;

		case ARRAY_REF:
		case ARRAY_RANGE_REF:
		case BIT_FIELD_REF:
		case COMPONENT_REF:
		case IMAGPART_EXPR:
		case REALPART_EXPR:
		case VIEW_CONVERT_EXPR:
			arg = TREE_OPERAND(arg, 0);
			break;
		}
	}

	return arg;
}

static inline tree get_inner_arg(tree arg)
{
	tree offset;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	return get_inner_reference(arg, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
}

static inline bool ptr_to_special_addr_space_p(const_tree arg)
{
	const_tree type = TREE_TYPE(arg);

	if (!POINTER_TYPE_P(type))
		return false;

	return !ADDR_SPACE_GENERIC_P(TYPE_ADDR_SPACE(TREE_TYPE(type)));
}

bool insn_dominates_insn_p(enum cdi_direction dir, rtx_insn *insn1, rtx_insn *insn2);

static inline struct opt_pass *plugin_find_pass(const char *name)
{
	int i;
	gcc::pass_manager *passes = g->get_passes();

	for (i = 0; i < passes->passes_by_id_size; i++) {
		struct opt_pass *pass = passes->get_pass_for_id(i);

		if (pass && !strcmp(pass->name, name))
			return pass;
	}

	return NULL;
}

static inline void set_current_function_decl(tree fndecl)
{
	gcc_assert(fndecl != NULL_TREE);

	push_cfun(DECL_STRUCT_FUNCTION(fndecl));
	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);
}

static inline void unset_current_function_decl(void)
{
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	pop_cfun();
}

// create the equivalent of
// asm volatile("" : : : "memory");
// or
// asm("" : "+rm"(var));
// or
// asm("" : "+rm"(var) : : "memory");
gimple_stmt_ptr gimple_barrier(tree var, bool full);

static inline bool single_arg_phi_p(gimple_stmt_ptr phi_stmt)
{
	if (gimple_code(phi_stmt) != GIMPLE_PHI)
		return false;
	return gimple_phi_num_args(phi_stmt) == 1;
}

bool tree_for_decl_and_aliases(tree decl, bool (*callback)(tree, void *), void *data);

basic_block get_single_pred_entry_block(void);

bool has_trait(tree decl, const char *trait);

static inline gcall *gimple_build_builtin_trap(location_t loc)
{
	gcall *g = as_a_gcall(gimple_build_call(builtin_decl_implicit(BUILT_IN_TRAP), 0));
	gimple_call_set_ctrl_altering(g, true);
	gimple_set_location(g, loc);
	return g;
}
#endif
