/*
 * Copyright 2022-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"
#include "gcc-ksyms.h"

static unsigned int rap_rethunk_execute(void)
{
	rtx_insn *insn;

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		rtx jump;

		if (INSN_DELETED_P(insn))
			continue;

		if (!returnjump_p(insn))
			continue;

		jump = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup("RETHUNK"), empty_string, 0, rtvec_alloc(0), rtvec_alloc(0), rtvec_alloc(0), INSN_LOCATION(insn));
		emit_insn_before(jump, insn);

		SET_INSN_DELETED(insn);
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	return 0;
}

static bool rap_rethunk_gate(void)
{
	if (has_trait(current_function_decl, "simpleret"))
		return false;

	if (MODULE != NULL_TREE)
		return true;

	if (has_trait(current_function_decl, "boot"))
		return false;

	return true;
}

#define PASS_NAME rap_rethunk
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_df_verify
#include "gcc-generate-rtl-pass.h"
