/*
 * Copyright 2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * align certain code sections to a higher value than the default 1
 */

#include "utilities.h"

unsigned int align_init_text_bits;

static const char * const align_text_sections[] = {
	".init.text",
	".exit.text",
	".meminit.text",
	".head.text",
	".ref.text",
};

static bool is_align_text_section(const char *name)
{
	size_t i;

	if (!name)
		return false;

	for (i = 0; i < ARRAY_SIZE(align_text_sections); i++)
		if (!strcmp(align_text_sections[i], name))
			return true;

	return false;
}

void align_init_text_sections(void *event_data, void *data __unused)
{
	tree fndecl = (tree)event_data;
	location_t loc;
	bool align_me;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	gcc_assert(TREE_CODE(TREE_TYPE(fndecl)) == FUNCTION_TYPE);
	loc = DECL_SOURCE_LOCATION(fndecl);

	/* we're done if the alignment is already what we want */
	if (DECL_ALIGN(fndecl) == align_init_text_bits)
		return;

	align_me = is_align_text_section(get_decl_section_name(fndecl));

	/* skip explicitly aligned functions */
	if (DECL_USER_ALIGN(fndecl)) {
		if (align_me)
			error_at(loc, "cannot align %qE to %qu bits, alignment manually set to %qu bits", fndecl, align_init_text_bits, DECL_ALIGN(fndecl));
		else if (0)
			inform(loc, "would not align %qE to %qu bits, alignment manually set to %qu bits", fndecl, align_init_text_bits, DECL_ALIGN(fndecl));
		return;
	}

	if (!align_me)
		return;

	/* we won't lower alignment */
	if (DECL_ALIGN(fndecl) > align_init_text_bits) {
		error_at(loc, "not lowering alignment of %qE", fndecl);
		return;
	}

//	inform(loc, "forcing alignment for %qE", fndecl);
	SET_DECL_ALIGN(fndecl, align_init_text_bits);
//	debug_tree(fndecl);
}
