/*
 * Copyright 2019-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * dump the current function into the gimple dump file as gcc 8 and newer do it
 */

#include "utilities.h"

#if BUILDING_GCC_VERSION < 8000
static void execute_function_dump(struct function *fn, void *data ATTRIBUTE_UNUSED)
{
	if (!dump_file)
		return;

	push_cfun(fn);

	if (fn->curr_properties & PROP_trees)
		dump_function_to_file(fn->decl, dump_file, dump_flags);
	else
		print_rtl_with_bb(dump_file, get_insns(), dump_flags);

	fflush(dump_file);

	if ((cfun->curr_properties & PROP_cfg) && (dump_flags & TDF_GRAPH))
		print_graph_cfg(dump_file_name, cfun);

	pop_cfun();
}

void (*old_internal_error)(diagnostic_context *, const char *, va_list *);

void emergency_dump_internal_error_function(diagnostic_context *dc, const char *msgid, va_list *ap)
{
	enum opt_pass_type pt;

	if (old_internal_error)
		old_internal_error(dc, msgid, ap);

	if (!current_pass || !dump_file || !cfun)
		return;

	pt = current_pass->type;
	fnotice(stderr, "during %s pass: %s\n", pt == GIMPLE_PASS ? "GIMPLE" : pt == RTL_PASS ? "RTL" : "IPA", current_pass->name);
	fnotice(stderr, "dump file: %s\n", dump_file_name);
	fprintf(dump_file, "\n\n\nEMERGENCY DUMP:\n\n");
	execute_function_dump(cfun, current_pass);
}
#endif
