#ifndef UTILITIES_H_INCLUDED
#define UTILITIES_H_INCLUDED

#include "gcc-common.h"

// align_init_text_sections
extern unsigned int align_init_text_bits;
void align_init_text_sections(void *event_data, void *data __unused);

// asm_include
extern const char *asm_include;
void asm_include_start_unit(void __unused *gcc_data, void __unused *user_data);

// context
extern struct attribute_spec context_attr;

// dead_copy_elimination
unsigned int dead_copy_elimination_execute(void);

// diagnose
extern struct attribute_spec diagnose_attr;
unsigned int diagnose_execute(void);

// emergency_dump
#if BUILDING_GCC_VERSION < 8000
extern void (*old_internal_error)(diagnostic_context *, const char *, va_list *);
void emergency_dump_internal_error_function(diagnostic_context *dc, const char *msgid, va_list *ap);
#endif

// fix_esra
unsigned int fix_esra_execute(void);

// fix_tailc
#ifdef TARGET_386
unsigned int fix_tailc_execute(void);
#endif

// ksyms
extern const struct ggc_root_tab gt_ggc_r_ksyms[];
void ksyms_find_decls(void *event_data, void *data __unused);

// late_inline
extern struct attribute_spec late_inline_attr;
void late_inline_register_hook(void);
#if BUILDING_GCC_VERSION >= 13000
void late_inline_delay_inlines(void *event_data __unused, void *data __unused);
#endif

// leaf_only
extern struct attribute_spec leaf_only_attr;

// maybe_used
extern struct attribute_spec maybe_used_attr;
unsigned int unmark_maybe_used_functions_execute(void);

// naked
#if defined(TARGET_386) && BUILDING_GCC_VERSION >= 6000 && BUILDING_GCC_VERSION < 8000
extern struct attribute_spec naked_attr;
void naked_start_unit(void __unused *gcc_data, void __unused *user_data);
#endif

// no_caller_saved_registers
#if defined(TARGET_386) && BUILDING_GCC_VERSION < 7000
extern struct attribute_spec no_caller_saved_registers_attr;
void no_caller_saved_registers_start_unit(void __unused *gcc_data, void __unused *user_data);
#endif

// noderef
extern struct attribute_spec noderef_attr;

// nolocal
extern bool report_nolocals;
extern struct attribute_spec nolocal_attr;
void nolocal_start_unit(void __unused *gcc_data, void __unused *user_data);
void nolocal_finish_type(void *event_data, void *data __unused);
void nolocal_typemap_allocate(void *event_data, void *data);
void nolocal_typemap_cleanup(void *event_data, void *data);
unsigned int convert_nolocal_variables_execute(void);

// no_sanitize
#if BUILDING_GCC_VERSION < 8000
extern struct attribute_spec no_sanitize_attr;
#endif
#if BUILDING_GCC_VERSION < 12000
extern struct attribute_spec no_sanitize_coverage_attr;
#endif
void no_sanitize_register_hook(void);

// no_stack_protector
extern struct attribute_spec no_stack_protector_attr;
void no_stack_protector_start(void *event_data, void *data __unused);
void no_stack_protector_end(void *event_data, void *data __unused);

// null_deref
extern bool report_null_deref;

// once
extern struct attribute_spec read_once_attr;
extern struct attribute_spec write_once_attr;
unsigned int once_after_execute(void);

// tweaked regparm(3)
#ifdef TARGET_386
void regparm_start_unit(void __unused *gcc_data, void __unused *user_data);
#endif

// section_type
extern struct attribute_spec section_type_attr;
void section_type_start_unit(void __unused *gcc_data, void __unused *user_data);

// traits
extern struct attribute_spec traits_attr;

// gcc passes
//simple_ipa_opt_pass *make_pass_colorize_rearm(gcc::context *);
gimple_opt_pass *make_pass_context(gcc::context *);
gimple_opt_pass *make_pass_diagnose(gcc::context *);
gimple_opt_pass *make_pass_early_discover_nolocal_variables(gcc::context *);
ipa_opt_pass_d *make_pass_ipa_discover_nolocal_variables(gcc::context *);
gimple_opt_pass *make_pass_convert_nolocal_variables(gcc::context *);
gimple_opt_pass *make_pass_dead_copy_elimination(gcc::context *);
gimple_opt_pass *make_pass_fix_esra(gcc::context *);
rtl_opt_pass *make_pass_fix_tailc(gcc::context *);
ipa_opt_pass_d *make_pass_late_inline(gcc::context *);
rtl_opt_pass *make_pass_leaf_only(gcc::context *);
#if defined(TARGET_386) && BUILDING_GCC_VERSION < 6000
rtl_opt_pass *make_pass_no_caller_saved_registers_gen_call(gcc::context *);
#endif
gimple_opt_pass *make_pass_no_sanitize_before_einline(gcc::context *);
gimple_opt_pass *make_pass_no_sanitize_after_einline(gcc::context *);
ipa_opt_pass_d *make_pass_no_sanitize_before_inline(gcc::context *);
ipa_opt_pass_d *make_pass_no_sanitize_after_inline(gcc::context *);
gimple_opt_pass *make_pass_once_before(gcc::context *);
gimple_opt_pass *make_pass_once_after(gcc::context *);
simple_ipa_opt_pass *make_pass_unmark_maybe_used_functions(gcc::context *);

#endif
